/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/configuration.h"
#include <cstdlib>                            // for getenv
#include <cassert>                            // for assert
#include <string>                             // for char_traits, string
#include "boost/iterator/iterator_facade.hpp" // for iterator_facade_base
#include "boost/json/impl/object.hpp"         // for object::at, object::begin
#include "boost/json/string.hpp"              // for string
#include "boost/json/string_view.hpp"         // for string_view
#include "caosdb/authentication.h"            // for PlainPasswordAuthentic...
#include "caosdb/connection.h"                // for TlsConnectionConfiguration
#include "caosdb/constants.h"                 // for LIBCAOSDB_CONFIGURATIO...
#include "caosdb/exceptions.h"                // for ConfigurationError

namespace caosdb::configuration {
using boost::filesystem::exists;
using boost::filesystem::path;
using boost::json::object;
using boost::json::value;
using caosdb::authentication::Authenticator;
using caosdb::authentication::PlainPasswordAuthenticator;
using caosdb::connection::CertificateProvider;
using caosdb::connection::ConnectionConfiguration;
using caosdb::connection::InsecureConnectionConfiguration;
using caosdb::connection::PemFileCertificateProvider;
using caosdb::connection::TlsConnectionConfiguration;
using caosdb::exceptions::ConfigurationError;
using caosdb::utility::get_home_directory;
using caosdb::utility::load_json_file;

auto ConnectionConfigurationHelper::CreateCertificateProvider(
  const object &from) const -> std::unique_ptr<CertificateProvider> {
  std::unique_ptr<CertificateProvider> certificate_provider;
  if (from.contains("server_certificate_path")) {
    const value &path_str = from.at("server_certificate_path");
    assert(path_str.is_string() == true);
    certificate_provider = std::make_unique<PemFileCertificateProvider>(
      std::string(path_str.as_string().c_str()));
  }
  return certificate_provider;
};

auto ConnectionConfigurationHelper::CreateAuthenticator(
  const object &from) const -> std::unique_ptr<Authenticator> {
  std::unique_ptr<Authenticator> authenticator;
  if (from.contains("authentication")) {
    assert(from.at("authentication").is_object());
    auto authentication = from.at("authentication").as_object();
    auto type = std::string("plain");
    if (authentication.contains("type")) {
      assert(authentication.at("type").is_string());
      type = std::string(authentication.at("type").as_string().c_str());
    }
    if (type == "plain") {
      assert(authentication.contains("username"));
      auto username = authentication.at("username");
      assert(username.is_string());

      assert(authentication.contains("password"));
      auto password = authentication.at("password");
      assert(password.is_string());

      authenticator = std::make_unique<PlainPasswordAuthenticator>(
        std::string(username.as_string().c_str()),
        std::string(password.as_string().c_str()));
    } else {
      throw ConfigurationError("Unknow authentication type: '" + type + "'.");
    }
  }
  return authenticator;
};

auto ConnectionConfigurationHelper::CreateConnectionConfiguration(
  const bool tls, const std::string &host, const int port,
  const CertificateProvider *certificate_provider,
  const Authenticator *authenticator) const
  -> std::unique_ptr<ConnectionConfiguration> {
  if (tls) {
    if (certificate_provider != nullptr && authenticator != nullptr) {
      // authenticated and special certificate
      return std::make_unique<TlsConnectionConfiguration>(
        host, port, *certificate_provider, *authenticator);
    } else if (certificate_provider != nullptr) {
      // unauthenticated, special certificate
      return std::make_unique<TlsConnectionConfiguration>(
        host, port, *certificate_provider);
    } else if (authenticator != nullptr) {
      // authenticated, no special certificate
      return std::make_unique<TlsConnectionConfiguration>(host, port,
                                                          *authenticator);
    }
    // unauthenticated, no special certificate
    return std::make_unique<TlsConnectionConfiguration>(host, port);
  } else {
    return std::make_unique<InsecureConnectionConfiguration>(host, port);
  }
};

auto ConnectionConfigurationHelper::IsTls(const object &from) const -> bool {
  bool tls = true;
  if (from.contains("tls")) {
    auto tls_switch = from.at("tls");
    assert(tls_switch.is_bool());
    tls = tls_switch.as_bool();
  }
  return tls;
};

auto ConnectionConfigurationHelper::CreateConnectionConfiguration(
  const object &from) const -> std::unique_ptr<ConnectionConfiguration> {
  assert(from.contains("host"));
  const auto &host = from.at("host");
  assert(host.is_string());

  assert(from.contains("port"));
  const auto &port = from.at("port");
  assert(port.is_int64());

  auto tls = IsTls(from);

  auto certificate_provider = CreateCertificateProvider(from);

  auto authenticator = CreateAuthenticator(from);

  return CreateConnectionConfiguration(
    tls, std::string(host.as_string().c_str()),
    static_cast<int>(port.as_int64()), certificate_provider.get(),
    authenticator.get());
};

auto ConfigurationManager::mReset() -> void {
  mClear();
  InitializeDefaults();
};

auto ConfigurationManager::mClear() -> void {
  json_configuration = value(nullptr);
  ConnectionManager::Reset();
}

auto ConfigurationManager::mLoadSingleJSONConfiguration(const path &json_file)
  -> void {
  if (!json_configuration.is_null()) {
    throw ConfigurationError("This CaosDB client has already been configured.");
  }
  if (!exists(json_file)) {
    throw ConfigurationError("Configuration file does not exist.");
  }

  json_configuration = load_json_file(json_file);
  // TODO(far future) validate against json-schema
};

auto ConfigurationManager::mGetConnectionConfiguration(
  const std::string &name) const -> std::unique_ptr<ConnectionConfiguration> {
  auto connection_json = GetConnection(name);
  return connection_configuration_helper.CreateConnectionConfiguration(
    connection_json);
};

auto ConfigurationManager::mGetDefaultConnectionName() const -> std::string {
  auto connections = GetConnections();
  if (connections.contains("default")) {
    auto default_connection = connections.at("default");
    if (default_connection.is_object()) {
      // the name is actually "default"
      return std::string("default");
    } else {
      assert(default_connection.is_string());
      auto default_connection_name = default_connection.as_string();
      // return the string value of connections.default
      return std::string(default_connection_name.c_str());
    }
  }
  if (connections.size() == 1) {
    // return the key of the first and only sub-element of connections.
    return connections.begin()->key().to_string();
  }
  throw ConfigurationError("Could not determine the default connection.");
};

auto ConfigurationManager::GetConfiguration() const -> const object & {
  if (json_configuration.is_null()) {
    throw ConfigurationError("This CaosDB client has not been configured.");
  }
  assert(json_configuration.is_object());
  return json_configuration.as_object();
};

auto ConfigurationManager::GetConnections() const -> const object & {
  const auto &configuration = GetConfiguration();
  if (!configuration.contains("connections")) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  const auto &connections_value = configuration.at("connections");
  if (connections_value.is_null()) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  assert(connections_value.is_object());
  const auto &connections_object = connections_value.as_object();
  if (connections_object.empty()) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  return connections_object;
};

auto ConfigurationManager::GetConnection(const std::string &name) const
  -> const object & {
  const auto &connections = GetConnections();
  if (connections.contains(name)) {
    const auto &result_connection = connections.at(name);
    assert(result_connection.is_object());
    return result_connection.as_object();
  }
  throw ConfigurationError("The connection '" + name +
                           "' has not been defined.");
};

auto ConfigurationManager::InitializeDefaults() -> void {

  // find the configuration file...
  std::unique_ptr<path> configuration_file_path;
  for (const std::string &configuration_file :
       caosdb::LIBCAOSDB_CONFIGURATION_FILES_PRECEDENCE) {
    if (configuration_file == "$CAOSDB_CLIENT_CONFIGURATION") {
      // user specified a file via the environment variable
      const auto *from_env_var = getenv("CAOSDB_CLIENT_CONFIGURATION");
      if (from_env_var != nullptr) {
        configuration_file_path = std::make_unique<path>(from_env_var);
        if (exists(*configuration_file_path)) {
          break;
        } else {
          configuration_file_path = nullptr;
          // TODO(tf) log warning: "Configuration file under
          // $CAOSDB_CLIENT_CONFIGURATION does not exist.
        }
      }
    } else {
      // check standard locations
      configuration_file_path = std::make_unique<path>();
      const path raw(configuration_file);
      // resolve home directory
      for (auto segment = raw.begin(); segment != raw.end(); ++segment) {
        if (segment->string() == "$HOME") {
          path expanded_home(get_home_directory());
          *configuration_file_path /= expanded_home;
        } else {
          *configuration_file_path /= *segment;
        }
      }
      if (exists(*configuration_file_path)) {
        break;
      } else {
        configuration_file_path = nullptr;
      }
    }
  }

  // ... and use the configuration file
  if (configuration_file_path != nullptr) {
    // TODO(tf): log which file has been used.
    mLoadSingleJSONConfiguration(*configuration_file_path);
  } else {
    // TODO(tf): log warning: no configuration files has been found
  }
}

} // namespace caosdb::configuration
