/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"
#include <grpcpp/create_channel.h>              // for CreateChannel
#include <grpcpp/impl/codegen/client_context.h> // for ClientContext
#include <grpcpp/impl/codegen/status.h>         // for Status
#include <grpcpp/security/credentials.h>        // for SslCredentials
#include <iostream>                             // for operator<<, basic_...
#include <stdexcept>                            // for runtime_error
#include <string>                               // for operator+, char_tr...
#include <memory>
#include "caosdb/authentication.h" // for Authenticator
#include "caosdb/configuration.h"
#include "caosdb/exceptions.h"                    // for AuthenticationError
#include "caosdb/info/v1alpha1/main.grpc.pb.h"    // for GeneralInfoService
#include "caosdb/info/v1alpha1/main.pb.h"         // for GetVersionInfoResp...
#include "caosdb/transaction.h"                   // for Transaction
#include "caosdb/utility.h"                       // for load_string_file
#include "caosdb/info.h"                          // for VersionInfo
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...

namespace caosdb::connection {
using caosdb::authentication::Authenticator;
using caosdb::configuration::ConfigurationManager;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::exceptions::AuthenticationError;
using caosdb::exceptions::ConnectionError;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::info::v1alpha1::GetVersionInfoRequest;
using caosdb::info::v1alpha1::GetVersionInfoResponse;
using caosdb::transaction::Transaction;
using caosdb::utility::load_string_file;
using grpc::InsecureChannelCredentials;
using grpc::SslCredentials;
using grpc::SslCredentialsOptions;

PemFileCertificateProvider::PemFileCertificateProvider(
  const std::string &path) {
  this->certificate_provider = load_string_file(path);
}

auto PemFileCertificateProvider::GetCertificatePem() const -> std::string {
  return this->certificate_provider;
}

PemCertificateProvider::PemCertificateProvider(
  const std::string &certificate_provider) {
  this->certificate_provider = certificate_provider;
}

auto PemCertificateProvider::GetCertificatePem() const -> std::string {
  return this->certificate_provider;
}

ConnectionConfiguration::ConnectionConfiguration(const std::string &host,
                                                 int port) {
  this->host = host;
  this->port = port;
}

auto ConnectionConfiguration::GetHost() const -> std::string {
  return this->host;
}

auto ConnectionConfiguration::GetPort() const -> int { return this->port; }

auto operator<<(std::ostream &out, const ConnectionConfiguration &configuration)
  -> std::ostream & {
  out << configuration.ToString();
  return out;
}

InsecureConnectionConfiguration::InsecureConnectionConfiguration(
  const std::string &host, int port)
  : ConnectionConfiguration(host, port) {
  this->credentials = InsecureChannelCredentials();
}

auto InsecureConnectionConfiguration::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto InsecureConnectionConfiguration::ToString() const -> std::string {
  return "InsecureConnectionConfiguration(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + ")";
}

TlsConnectionConfiguration::TlsConnectionConfiguration(const std::string &host,
                                                       int port)
  : ConnectionConfiguration(host, port) {
  SslCredentialsOptions options;
  this->credentials = SslCredentials(options);
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port,
  const CertificateProvider &certificate_provider)
  : ConnectionConfiguration(host, port) {
  SslCredentialsOptions options;
  options.pem_root_certs = certificate_provider.GetCertificatePem();
  this->credentials = SslCredentials(options);
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port, const Authenticator &authenticator)
  : ConnectionConfiguration(host, port) {

  SslCredentialsOptions options;
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port,
  const CertificateProvider &certificate_provider,
  const Authenticator &authenticator)
  : ConnectionConfiguration(host, port) {

  SslCredentialsOptions options;
  options.pem_root_certs = certificate_provider.GetCertificatePem();
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

auto TlsConnectionConfiguration::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto TlsConnectionConfiguration::ToString() const -> std::string {
  return "TlsConnectionConfiguration(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + "," + this->certificate_provider +
         ")";
}

Connection::Connection(const ConnectionConfiguration &configuration) {
  const std::string target =
    configuration.GetHost() + ":" + std::to_string(configuration.GetPort());
  this->channel =
    grpc::CreateChannel(target, configuration.GetChannelCredentials());
  this->general_info_service = GeneralInfoService::NewStub(this->channel);
  this->entity_transaction_service =
    std::make_shared<EntityTransactionService::Stub>(this->channel);
}

auto operator<<(std::ostream &out, const Connection & /*connection*/)
  -> std::ostream & {
  out << "Connection()";
  return out;
}

[[nodiscard]] auto Connection::GetVersionInfo() const
  -> std::unique_ptr<VersionInfo> {
  const GetVersionInfoRequest request;
  GetVersionInfoResponse response;
  grpc::ClientContext context;
  const grpc::Status status =
    this->general_info_service->GetVersionInfo(&context, request, &response);

  if (!status.ok()) {
    switch (status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      throw AuthenticationError(status.error_message());
    case grpc::StatusCode::UNAVAILABLE:
      throw ConnectionError(status.error_message());
    default:
      throw std::runtime_error("Status Code " +
                               std::to_string(status.error_code()) + " - " +
                               status.error_message());
    }
  }
  return std::make_unique<VersionInfo>(response.release_version_info());
}

[[nodiscard]] auto Connection::CreateTransaction() const
  -> std::unique_ptr<Transaction> {
  auto service_stub = this->entity_transaction_service;
  return std::make_unique<Transaction>(service_stub);
};

auto ConnectionManager::mHasConnection(const std::string &name) const -> bool {
  auto it = connections.find(name);
  return it != connections.end();
}

auto ConnectionManager::mGetConnection(const std::string &name) const
  -> const std::shared_ptr<Connection> & {
  if (!HasConnection(name)) {
    try {
      auto connection = ConfigurationManager::GetConnectionConfiguration(name);
      connections[name] = std::make_shared<Connection>(*connection.release());
    } catch (const caosdb::exceptions::ConfigurationError &exc) {
      throw caosdb::exceptions::UnknownConnectionError("No connection named '" +
                                                       name + "' present.");
    }
  }
  return this->connections.at(name);
}

auto ConnectionManager::mGetDefaultConnection() const
  -> const std::shared_ptr<Connection> & {
  if (!HasConnection(default_connection_name)) {
    default_connection_name = ConfigurationManager::GetDefaultConnectionName();
    auto default_connection =
      ConfigurationManager::GetDefaultConnectionConfiguration();
    connections[default_connection_name] =
      std::make_shared<Connection>(*default_connection.release());
  }

  return connections.at(default_connection_name);
}

} // namespace caosdb::connection
